/*
 *  Jomar3D, a lobster-running-to-catch game.
 *  Copyright (C) 2006-2008 Jonathan Michalon <studios.chalmion@no-log.org>.
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <ctype.h>
#include <stdlib.h>

#include <GL/glut.h>

#include "core.h"
#include "display.h"
#include "draw.h"
#include "drawmenu.h"
#include "interface.h"
#include "jomar3D.h"
#include "scores.h"
#include "sound.h"
#include "utils.h"


static void interface_doMenu (void);
static void interface_rotateMenu (int dummy);
static void interface_changeOption (int opt);

/* interface_rotateMenu static vars */
static int l_beginrot = 0;    /* time at which the rotation begun, 0 if not rotating */
static int l_cylnb = -1;      /* which cylinder is turning */
static int l_dochange = 1;    /* Whether we should do the change or have already done it */

static int l_uppressed;       /* remember if the up key is pressed */
static Coord2i l_move_begin;  /* When moving the mouse, begin of the move */ 

static int l_object = 0;      /* current object selected */

static int l_ngamers;         /* number of gamers in the score file */



/* process the user's keyboard actions */
void
interface_onKeyDown (unsigned char c, int x, int y)
{
  int i;
  (void) x, (void) y;

  if (core_checkState (JOMAR3D_INMENU) && isdigit (c))
  { /* emulate that pressed number keys are as if an object was selected in the menu */
    l_object = atoi ((char *) &c);
    interface_doMenu ();
    return;
  }

/*  the first letter of each cylinder in the menu must enter this item - and stop all other action */
  if (core_checkState (JOMAR3D_INMENU))
  {
    for (i = 0; i < NCYL; i++)
    {
      if (tolower (c) == drawmenu_getCylinderLetter (i))
      {
        debugmsg (("cylinder number %d met by key pressed\n", i + 1));
        
        l_object = i + 1;
        interface_doMenu ();
        return;
      }
    }
  }

  switch (c)
  {
    case 27:                   /* Escape */
      exit (EXIT_SUCCESS);
      break;
    case 'm':                  /* return into main menu */
      interface_changeMenu (MP);
      break;
#if (USESOUND)
    case 'c':                  /* change music */
      sound_changeMusic ();
      break;
#endif
    case 'w':
      glutSwapBuffers ();
      break;
    case 'v':                  /* reposition view */
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
      /* revert v to a number character */
      if (c == 'v') c = '1';
      draw_setView (c - '0');
      glutPostRedisplay ();
      break;
    case 'f':               /* toggle game mode (full screen) */
    case 'F':
      display_switchMode ();
      break;
    case 'a':               /* try to eat */
    case 'A':
    case 13:
      core_tryEat ();
      break;
    case 'b':               /* boosting the lobster's speed */
    case 'B':
      core_startBoost ();
      break;
#if (USESOUND)
    case '+':
      sound_changeVolume (S_INCREASE, 0.);
      break;
    case '-':
      sound_changeVolume (S_DECREASE, 0.);
      break;
    case ' ':
      sound_pauseResume ();
      /* update menu items */
      drawmenu_createLists (0);
      drawmenu_redraw ();
      glutPostRedisplay ();
      break;
#endif /* (USESOUND) */
    default:
      debugmsg (("unused key %c (code %d)\n", c, c));
      break;
  }
}

/* Take care that a Special key aws pressed */
void
interface_onSKeyDown (int c, int x, int y)
{
  (void) x, (void) y;

  if (core_getLost ())
    return;

  switch (c)
  {
    case GLUT_KEY_RIGHT:
      core_wantAction (2, c);
      break;
    case GLUT_KEY_DOWN:
      core_wantAction (4, c);
      break;
    case GLUT_KEY_LEFT:
      core_wantAction (3, c);
      break;
    case GLUT_KEY_UP:
      core_wantAction (1, c);
      l_uppressed = 1;
      break;
    case GLUT_KEY_F11:
      interface_onKeyDown ('f', 0, 0);
      return;
      break;
    case GLUT_KEY_F10:
      interface_onKeyDown ('m', 0, 0);
      return;
      break;
    default:
      debugmsg (("unused key : %c %d\n", c, c));
      break;
  }

/* notice the motion */

  if (c == GLUT_KEY_RIGHT || c == GLUT_KEY_LEFT ||
      c == GLUT_KEY_UP    || c == GLUT_KEY_DOWN)
      core_startMotion ();
}

/* Take attention that a key was released */
void
interface_onKeyUp (int c, int x, int y)
{
  (void) x, (void) y;
  
  switch (c)
  {
    case GLUT_KEY_UP:
      l_uppressed = 0;
    case GLUT_KEY_RIGHT:
    case GLUT_KEY_LEFT:
    case GLUT_KEY_DOWN:
      core_stopMotion ();
      if (l_uppressed)
        interface_onSKeyDown (GLUT_KEY_UP, 1, 1);   /* restart if we had turn */
  }
}

/* manipulate the menus */
void
interface_changeMenu (int menu)
{
  if (!core_checkState (JOMAR3D_INMENU))
    core_initGame ();
  /* reset all menus */
  core_deleteState (JOMAR3D_MENU_MAIN | JOMAR3D_MENU_RULES | JOMAR3D_MENU_FINDS |
            JOMAR3D_MENU_DISPSCORE | JOMAR3D_MENU_OPTION);

  /* say we aren't in the scores (can come from there) */
  core_deleteState (JOMAR3D_SCORE);

  /* set the inmenu state */
  if (menu == NOMENU)
    core_deleteState (JOMAR3D_INMENU);
  else
    core_addState (JOMAR3D_INMENU);


  /* set the current menu */
  switch (menu)
  {
    case MP:
      core_addState (JOMAR3D_MENU_MAIN);
      break;
    case RULES:
      core_addState (JOMAR3D_MENU_RULES);
      break;
    case SCORES:
      core_addState (JOMAR3D_MENU_DISPSCORE);
      break;
    case OPTIONS:
      core_addState (JOMAR3D_MENU_OPTION);
      break;
    case FINDS:
      core_addState (JOMAR3D_MENU_FINDS);
      break;
  }

  drawmenu_setMenuNumber (menu);

  drawmenu_createLists (0);
  l_object = 0;
  draw_redraw ();                    /* resolves a lot of problems with menus when there are two buffers AND one for the window */
  draw_redraw ();
  glutPostRedisplay ();
}

/* Do the action of the menu according to the cylinder hit */
static void
interface_doMenu (void)
{
  if (core_checkState (JOMAR3D_MENU_MAIN) && l_object == 1)
    core_initGame ();
  if (core_checkState (JOMAR3D_MENU_MAIN) && l_object == 2)
    interface_changeMenu (RULES);
  if (core_checkState (JOMAR3D_MENU_MAIN) && l_object == 3)
  {
    l_ngamers = scores_load ();
    drawmenu_setNameToSee (-1);
    interface_changeMenu (SCORES);
    return;               /* So we don't clic 3rd cylinder directly */
  }
  if (core_checkState (JOMAR3D_MENU_MAIN) && l_object == 4)
  {
    interface_changeMenu (OPTIONS);
    return;               /* So we don't do the opposit a few lines after there */
  }
  if (core_checkState (JOMAR3D_MENU_MAIN) && l_object == 5)
    interface_changeMenu (FINDS);
  if (core_checkState (JOMAR3D_MENU_RULES) && l_object == 1)
    interface_changeMenu (MP);
  if (core_checkState (JOMAR3D_MENU_DISPSCORE))
  {
    if (l_object > l_ngamers)
    {
      interface_changeMenu (MP);
      return;
    }
    drawmenu_setNameToSee (l_object - 1);
    draw_redraw ();
    glutPostRedisplay ();
  }
  if (core_checkState (JOMAR3D_MENU_OPTION) && l_object == 5)
    interface_changeMenu (MP);
  if (core_checkState (JOMAR3D_MENU_OPTION) &&
      (l_object >= 1 && l_object <= 4))
    if (!l_beginrot)
      interface_rotateMenu (0);
  if (core_checkState (JOMAR3D_MENU_FINDS) && l_object == 1)
    interface_changeMenu (MP);
}

/* process the mouse buttons' actions */
void
interface_onMouse (int button, int state, int x, int y)
{
  if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN)
  {
    core_addState (JOMAR3D_SHADING);
    l_move_begin.x = x;
    l_move_begin.y = y;
  }
  if (button == GLUT_LEFT_BUTTON && state == GLUT_UP)
  {
    core_deleteState (JOMAR3D_SHADING);
  }

  if (button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN)
  {
    core_addState (JOMAR3D_MOVING_H);
    l_move_begin.x = x;
    l_move_begin.y = y;
  }
  if (button == GLUT_RIGHT_BUTTON && state == GLUT_UP)
  {
    core_deleteState (JOMAR3D_MOVING_H);
  }
  if (button == GLUT_MIDDLE_BUTTON && state == GLUT_DOWN)
  {
    core_addState (JOMAR3D_MOVING_V);
    l_move_begin.x = x;
    l_move_begin.y = y;
  }
  if (button == GLUT_MIDDLE_BUTTON && state == GLUT_UP)
  {
    core_deleteState (JOMAR3D_MOVING_V);
  }

  if (l_object == 0 && (button == 3 || button == 4))
  {
    if (button == 3)
      draw_rotateView (1);
    else if (button == 4)
      draw_rotateView (-1);
    glutPostRedisplay ();
  }

  if (state == GLUT_DOWN && l_object != 0 &&
      button != GLUT_MIDDLE_BUTTON && !l_beginrot)
  {
    interface_doMenu ();
    if (button == GLUT_RIGHT_BUTTON || button == 3)
      core_addState (JOMAR3D_TURN_BACK);
  }
}

/* Process the move of the mouse when passive (button pressed) */
void
interface_motionActive (int x, int y)
{
  float dx = (x - l_move_begin.x) / 4.;
  float dy = (y - l_move_begin.y) / 4.;

  glutSetCursor (GLUT_CURSOR_CYCLE);

  if (core_checkState (JOMAR3D_SHADING))
      draw_shade (dx, dy);
  if (core_checkState (JOMAR3D_MOVING_H))
      draw_moveViewH (dx, dy);
  if (core_checkState (JOMAR3D_MOVING_V))
      draw_moveViewV (dx, dy);

  l_move_begin.x = x;
  l_move_begin.y = y;

  glutPostRedisplay ();
}

/* change an option from the menu and remember it in the conf file */
static void
interface_changeOption (int opt)
{
  debugmsg (("Changing option number %d\n", opt));
  
  switch (opt)
  {
    case 0:
      if (core_checkState (JOMAR3D_TURN_BACK))
        core_setLanguage (-1);
      else
        core_setLanguage ( 1);
      break;
    
    case 1:
      if (core_checkState (JOMAR3D_TURN_BACK))
        core_setSkin (-1);
      else
        core_setSkin ( 1);
      break;
    
#if (USESOUND)
    case 2:
      sound_pauseResume ();
      drawmenu_createLists (0);
      draw_redraw ();
      core_saveOptions ();
      glutPostRedisplay ();
      break;
#endif
    
    case 3:
      if (core_checkState (JOMAR3D_TURN_BACK))
        draw_applyQuality (1, 1);
      else
        draw_applyQuality (1, -1);
      drawmenu_createLists (0);
      draw_redraw ();
      glutPostRedisplay ();
      break;
  }
  
  core_saveOptions ();
}

/* rotate a cylinder from the menu, changing an option */
static void
interface_rotateMenu (int dummy)
{
  float elapsed, cangle;
#define ROTTIME 1000.
  
  (void) dummy;
  if (l_beginrot == 0)
    l_beginrot = glutGet (GLUT_ELAPSED_TIME) - 1; /* -1: so the cylinder moves already now (don't accept more event) */
  elapsed = glutGet (GLUT_ELAPSED_TIME) - l_beginrot;

  if (l_cylnb == -1)
    l_cylnb = l_object - 1;
  if (elapsed > ROTTIME)
  {
    drawmenu_setCylinderAngle(l_cylnb, 0.);
    l_beginrot = 0;
    l_cylnb = -1;
    l_dochange = 1;
    core_deleteState (JOMAR3D_TURN_BACK);
    goto draw;
  }

  cangle = elapsed / ROTTIME * 2. * 180.;
  if (core_checkState (JOMAR3D_TURN_BACK))
    cangle = -cangle;
  drawmenu_setCylinderAngle (l_cylnb, cangle);

  glutTimerFunc (50., interface_rotateMenu, 0);

  if (l_dochange && elapsed > ROTTIME / 2.)
  {
    interface_changeOption (l_cylnb);
    l_dochange = 0;
  }

#undef ROTTIME
  
draw:;
  drawmenu_createLists (l_cylnb + 1);
  drawmenu_redraw ();
}


/* change the color of a cylinder (the mouse is on this cylinder) */
static void
interface_colorateMenu (int sobject)
{
  if (!core_checkState (JOMAR3D_INMENU))
    return;
  if (l_object == 0)
  {
    drawmenu_redraw ();
    return;
  }
  if (l_object != sobject)
    return;

  drawmenu_drawHitCylender (l_object);

/* prepare the next redisplay */
  glutTimerFunc (50., interface_colorateMenu, l_object);
}

/* Process the action of the mouse, no button pressed when called (passive) */
void
interface_motionPassive (int x, int y)
{
  GLint hits;
  GLuint depth = (GLuint) ~ 0;
  unsigned int getThisName;
  GLuint names, *ptr;
  GLint i;
  int lastobject = l_object;

  glutSetCursor (GLUT_CURSOR_LEFT_ARROW);

  l_object = 0;
  if (!core_checkState (JOMAR3D_INMENU))
    return;

/* when in the menu, do a special rendering to know if a cylinder is pointed */
  
/* select */
  ptr = draw_pickingRender (x, y);
  hits = glRenderMode (GL_RENDER);
  
  if (hits == 0)
    return;

/* process results of selecting */
  for (i = 0; i < hits; i++)
  {                           /* for each hit  */
    getThisName = 0;
    names = *ptr;
    ptr++;                    /* skip # name */
    if (*ptr <= depth)
      {
        depth = *ptr;
        getThisName = 1;
      }
    ptr++;                    /* skip z1 */
    if (*ptr <= depth)
      {
        depth = *ptr;
        getThisName = 1;
      }
    ptr++;                    /* skip z2 */

    if (getThisName)
      l_object = *ptr;
    ptr += names;             /* skip the names list */
  }
  
  debugmsg (("Selected object: %d at position: %i %i\n", l_object, x, y));
  
  if (l_object == -1)
  {
    l_object = 0;
    return;
  }

  glutSetCursor (GLUT_CURSOR_INFO);

  if (l_object != lastobject)
  {
    interface_colorateMenu (l_object);
#if (USESOUND)
    glutTimerFunc (10, sound_mix, SMENU1 + l_object - 1);
#endif
    /*glutPostRedisplay ();*/
  }
}
