/*
 *  readpng, a simple interface to load PNG to an OpenGL compatible structure.
 *  Copyright (C) 2008  Colomban Wendling <ban@herbesfolles.org>
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/*
 * Based on the tutorial from David HENRY available on Developpez.com:
 * <http://tfcduke.developpez.com/tutoriel/format/png/>
 */

/*
gcc -pipe -Wall -W -O2 -Wunused -Wunreachable-code -g -ansi -pedantic \
    readpngtest.c readpng.c \
    $(libpng12-config --libs --cflags) -fPIC -lGL -lGLU -lglut \
    -o png.out
 */

/*
 * creation date: Tue Sep  9 17:00:00 CEST 2008
 * revision date: Sat Sep 13 22:06:48 CEST 2008
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h> /* in hope to detect machine's bitwise */
#include <GL/gl.h>
#include <GL/glu.h>
#include <png.h>
#include "readpng.h"


/* error/warning messages prefixes */
#define READPNG_EEPRFX "**readpng: "
#define READPNG_WWPRFX " *readpng: "


/** Creates a new RPNGTextureGL structure
 * @returns a pointer to the newly allocated RPNGTextureGL that should be freed
 *          later or NULL on failure
 * 
 * @see readpng_tgl_free()
 */
RPNGTextureGL* readpng_tgl_new (void)
{
  RPNGTextureGL* tgl;
  
  tgl = malloc (sizeof *tgl);
  if (tgl)
  {
    tgl->width          = 0;
    tgl->height         = 0;
    tgl->format         = 0;
    tgl->internalFormat = 0;
    tgl->texels         = NULL;
  }
  
  return tgl;
}

/** Frees a previously allocated RPNGTextureGL
 * @param tgl   the RPNGTextureGL to free
 */
void readpng_tgl_free (RPNGTextureGL* tgl)
{
  if (tgl)
  {
    if (tgl->texels)
      free (tgl->texels);
    free (tgl);
  }
}


/** Creates a new RPNGBuffer structure
 * @param free_func   the function to use to free buffer->data or NULL to use
 *                    the standard C free() function
 * @returns a pointer to the newly allocated RPNGBuffer that should be freed
 *          later or NULL on failure
 * 
 * @see readpng_buf_free()
 */
RPNGBuffer* readpng_buf_new (RPNGDataFreeFunc free_func)
{
  RPNGBuffer* buf;
  
  buf = malloc (sizeof *buf);
  if (buf)
  {
    buf->data = NULL;
    buf->length = 0;
    buf->offset = 0;
    
    /* function to free data */
    if (free_func)
      buf->data_free_func = free_func;
    /* default free func is free() */
    else
      buf->data_free_func = free;
  }
  
  return buf;
}

/** Frees a previously allocated RPNGBuffer
 * @param buf   the RPNGBuffer to free
 */
void readpng_buf_free (RPNGBuffer* buf)
{
  if (buf)
  {
    if (buf->data)
      buf->data_free_func (buf->data);
    free (buf);
  }
}

/** Creates a new RPNGBuffer structure and loads @p fp data in
 * @param fp  a file stream from where read data
 * @returns a pointer to the newly allocated RPNGBuffer containing @p fp data or
 *          NULL on failure
 */
RPNGBuffer* readpng_buf_new_from_stream (FILE* fp)
{
  RPNGBuffer* buf = NULL;
  
  buf = readpng_buf_new (free);
  if (!buf)
  {
    perror (READPNG_EEPRFX"readpng_buf_new()");
    return NULL;
  }
  
  /* get file length */
  fseek (fp, 0, SEEK_END);
  buf->length = ftell (fp);
  fseek (fp, 0, SEEK_SET);
  buf->offset = 0;
  
  buf->data = malloc (buf->length);
  if (!buf->data)
  {
    perror (READPNG_EEPRFX"malloc()");
    readpng_buf_free (buf);
    return NULL;
  }
  
  /* read whole file data */
  fread (buf->data, 1, buf->length, fp);
  
  return buf;
}


/** Sets fromat of a RPNGTextureGL from libpng's data enum
 * @param tgl           a RPNGTextureGL
 * @param pngcolortype  libpng's color type
 * @returns 0 if @p pngcolortype is known and supported (e.g. on success) and
 *          -1 otherwise
 * 
 * @internal
 */
static int readpng_tgl_set_format (RPNGTextureGL* tgl, int pngcolortype)
{
  if (!tgl)
    return -1;
  
  switch (pngcolortype)
  {
    case PNG_COLOR_TYPE_GRAY:
      tgl->format = GL_LUMINANCE;
      tgl->internalFormat = 1;
      break;

    case PNG_COLOR_TYPE_GRAY_ALPHA:
      tgl->format = GL_LUMINANCE_ALPHA;
      tgl->internalFormat = 2;
      break;

    case PNG_COLOR_TYPE_RGB:
      tgl->format = GL_RGB;
      tgl->internalFormat = 3;
      break;

    case PNG_COLOR_TYPE_RGB_ALPHA:
      tgl->format = GL_RGBA;
      tgl->internalFormat = 4;
      break;

    default:
      /* badness */
      fprintf (stderr, READPNG_EEPRFX"bad color type %d\n", pngcolortype);
      return -1;
  }
  
  return 0;
}

/* callback to read data from a RPNGBuffer for libpng */
static void readpng_read_from_buf (png_structp png_ptr,
                                   png_bytep data,
                                   png_size_t length)
{
  RPNGBuffer *src = (RPNGBuffer*)png_get_io_ptr (png_ptr);
  
  /* check if read request length is no more than buffer length */
  if ((unsigned long)src->length < src->offset + length)
  {
    /* invalid read request, no data left (probably truncated PNG) */
    
    fprintf (stderr, READPNG_EEPRFX"bad read request: no data left in buffer\n");
    /* avoid BOF, read no more than buffer length */
    /* unused because of png_error() below */
    /*length = src->length - src->offset;*/
    
    /* raise an error, we can't continue without data */
    png_error (png_ptr, "Unexpected end of data");
  }
  
  /* copy data from image buffer */
  memcpy (data, src->data + src->offset, length);
  
  /* advance in the file */
  src->offset += length;
}


/** Fills a texture with a PNG buffer
 * @param tgl     a RPNGTextureGL texture to fill
 * @param buf     RPNGBuffer buffer from where read PNG data
 * @param pngp    an initialised libpng png_structp
 * @param infop   the libpng png_infop corresponding to @p pngp
 * @returns 0 on success or another value on failure
 * 
 * This function does the real PNG read with libpng
 */
static int readpng_tgl_set_from_png (RPNGTextureGL* tgl,
                                     RPNGBuffer* buf,
                                     png_structp pngp,
                                     png_infop infop)
{
  int                   colordepth;
  int                   colortype;
  png_bytep* volatile   row_pointers = NULL; /* volatile for use on setjmp */
  unsigned int          i;
  /* png_uint_32 is 64 bit on some architectures!
   * then we need real png_uint_32 as temporary variables */
  png_uint_32     tmpwidth32;
  png_uint_32     tmpheight32;
  
  /* define to cleanly abort the function */
  #define CLEANRET(v)\
    do { \
      free (row_pointers), row_pointers = NULL; \
      return (v); \
    } while (0)
  
  /* initialize the setjmp to return properly after a libpng error occured */
  if (setjmp (png_jmpbuf (pngp)))
  {
    CLEANRET (1);
  }
  
  /* We can now safely start to work with libpng */
  
  /* set "png_read" callback function and give source of data */
  png_set_read_fn (pngp, (png_voidp*)buf, readpng_read_from_buf);
  
  /* read png info */
  png_read_info (pngp, infop);
  
  /* retrive depth & color type */
  colordepth = png_get_bit_depth (pngp, infop);
  colortype  = png_get_color_type (pngp, infop);
  
  /* request converting indexed color images to RGB images */
  if (colortype == PNG_COLOR_TYPE_PALETTE)
    png_set_palette_to_rgb (pngp);
  
  /* convert 1-2-4 bits grayscale images to 8 bits grayscale. */
  if (colortype == PNG_COLOR_TYPE_GRAY && colordepth < 8)
    png_set_gray_1_2_4_to_8 (pngp);

  if (png_get_valid (pngp, infop, PNG_INFO_tRNS))
    png_set_tRNS_to_alpha (pngp);
  
  /* if needed, request unpacking of pixels */
  if (colordepth == 16)
    png_set_strip_16 (pngp);
  else if (colordepth < 8)
    png_set_packing (pngp);
  
  /* update info structure to apply transformations */
  png_read_update_info (pngp, infop);
  
  /* retrieve updated information */
  /* png_uint_32 is 64bits under certain architectures, then we need to store it
   * in a real png_uint_32 variable before copying it to GLuint height and width
   * of our structure. */
  png_get_IHDR (pngp, infop,
                &tmpwidth32, &tmpheight32,
                &colordepth, &colortype,
                NULL, NULL, NULL);
  /* if on 64 bits or unknown, hack */
  #if !defined (__WORDSIZE) || __WORDSIZE > 32
  /* test if width or height fill into OpenGL's GLuint */
  if (tmpwidth32 > ((GLuint)~0U) || tmpheight32 > ((GLuint)~0U))
  {
    fprintf (stderr, READPNG_EEPRFX"image too big (%ldx%ld)\n",
             tmpwidth32, tmpheight32);
    CLEANRET (1);
  }
  #endif
  tgl->width  = (GLuint)tmpwidth32;
  tgl->height = (GLuint)tmpheight32;
  
  /* set TGL format from PNG one */
  if (readpng_tgl_set_format (tgl, colortype) != 0)
    CLEANRET (1);
  
  /* allocate memory for pixel data */
  tgl->texels = malloc (sizeof (*tgl->texels) * tgl->width * tgl->height *
                        tgl->internalFormat);
  
  /* map correctly pixels for OpenGL */
  /* setup a pointer array. Each one points at the begening of a row. */
  row_pointers = malloc (sizeof (*row_pointers) * tgl->height);
  
  for (i = 0; i < tgl->height; i++)
  {
    row_pointers[i] = (png_bytep)(tgl->texels +
                                  ((tgl->height - (i + 1)) *
                                   tgl->width * tgl->internalFormat));
    /*
    row_pointers[i] = (png_bytep)(tgl->texels + (i*tgl->width * tgl->internalFormat));
    */
  }
  /* read pixel data using row pointers */
  png_read_image (pngp, row_pointers);
  /* we don't need row pointers anymore */
  free (row_pointers), row_pointers = NULL;
  
  /* finish decompression */
  png_read_end (pngp, NULL);
  
  
  #undef CLEANRET
  
  
  return 0;
}

/** Creates a new RPNGTextureGL from a buffer
 * @param buf   a RPNGBuffer from where read data
 * @returns a newl allocated RPNGTextureGL on success or NULL on failure
 */
RPNGTextureGL* readpng_tgl_new_from_buffer (RPNGBuffer* buf)
{
  png_structp     pngp;
  png_infop       infop;
  RPNGTextureGL*  tgl = NULL;
  
  /* check for valid magic number */
  if (!png_check_sig (buf->data, 8))
  {
    fprintf (stderr, READPNG_EEPRFX"invalid PNG: bad magic\n");
    return NULL;
  }
  
  /* create a png read struct */
  pngp = png_create_read_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (!pngp)
    return NULL;
  
  /* create a png info struct */
  infop = png_create_info_struct (pngp);
  if (!infop)
  {
    png_destroy_read_struct (&pngp, NULL, NULL);
    return NULL;
  }
  
  /* our personal struct */
  tgl = readpng_tgl_new ();
  if (!tgl)
  {
    perror (READPNG_EEPRFX"malloc()");
    png_destroy_read_struct (&pngp, &infop, NULL);
    return NULL;
  }
  
  /* We can now safely start to work with libpng */
  
  /* reads PNG data and fill our texture */
  if (readpng_tgl_set_from_png (tgl, buf, pngp, infop) != 0)
  {
    /* if read fails, free texture and set it to NULL to safely return it */
    readpng_tgl_free (tgl), tgl = NULL;
  }
  
  /* release memory */
  png_destroy_read_struct (&pngp, &infop, NULL);
  
  return tgl;
}

/** Creates a new RPNGTextureGL from a file stream
 * @param fp  a file stream from where read data
 * @returns a newl allocated RPNGTextureGL on success or NULL on failure
 * 
 * @see readpng_tgl_new_from_buffer
 * @see readpng_tgl_new_from_file
 */
RPNGTextureGL* readpng_tgl_new_from_stream (FILE* fp)
{
  RPNGTextureGL* tgl = NULL;
  RPNGBuffer*    buf = NULL;
  
  buf = readpng_buf_new_from_stream (fp);
  if (!buf)
  {
    fprintf (stderr, READPNG_EEPRFX"failed to create buffer\n");
    return NULL;
  }
  
  /* now launch loading */
  tgl = readpng_tgl_new_from_buffer (buf);
  
  readpng_buf_free (buf);
  
  return tgl;
}

/** Creates a new RPNGTextureGL from a file
 * @param filename  a filename from where try to read data
 * @returns a newl allocated RPNGTextureGL on success or NULL on failure
 * 
 * @see readpng_tgl_new_from_buffer
 * @see readpng_tgl_new_from_stream
 */
RPNGTextureGL* readpng_tgl_new_from_file (const char* filename)
{
  FILE* fp;
  RPNGTextureGL* tgl;
  
  fp = fopen (filename, "rb");
  if (!fp)
  {
    perror (READPNG_EEPRFX"fopen()");
    return NULL;
  }
  
  if (!(tgl = readpng_tgl_new_from_stream (fp)))
    fprintf (stderr, READPNG_EEPRFX"failed to load PNG file '%s' (see above)\n",
             filename);
  
  fclose (fp);
  
  return tgl;
}


/*=============================== OpenGL part ================================*/

/** Loads a texture into a new OpenGL's texture slot
 * @param filename  a filename from where try to load the texture
 * @returns The new OpenGL's texture ID or 0 on failure
 * 
 * This function creates an OpenGL texture from the given file.
 * 
 * @note Please note that this function binds the new texture on success.
 */
GLuint readpng_load_texture_gl (const char* filename)
{
  RPNGTextureGL* tgl;
  GLuint texid = 0;
  
  tgl = readpng_tgl_new_from_file (filename);
  
  if (tgl)
  {
    /* OpenGL don't like odd textures... */
    if (tgl->width & 0x01 || tgl->height & 0x01)
    {
      fprintf (stderr, READPNG_EEPRFX"texture have odd dimentions (%dx%d)\n",
               tgl->width, tgl->height);
    }
    else
    {
      /* generate texture */
      glGenTextures (1, &texid);
      glBindTexture (GL_TEXTURE_2D, texid);
    
      /* setup some parameters for texture filters and mipmapping */
      glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      
      glTexImage2D (GL_TEXTURE_2D, 0, tgl->internalFormat,
                    tgl->width, tgl->height, 0, tgl->format,
                    GL_UNSIGNED_BYTE, tgl->texels);
    }
    
    /* OpenGL has its own copy of texture data; we can safely delete our own */
    readpng_tgl_free (tgl);
  }
  
  return texid;
}

/** Loads texture mipmaps into a new OpenGL's texture slot
 * @param filename  a filename from where try to load the texture
 * @returns The new OpenGL's texture ID or 0 on failure
 * 
 * This function generates an OpenGL texture mipmaps from the given file.
 * 
 * @note Please note that this function binds the new texture on success.
 */
GLuint readpng_load_mipmaps_gl (const char* filename)
{
  RPNGTextureGL* tgl;
  GLuint texid = 0;
  
  tgl = readpng_tgl_new_from_file (filename);
  
  if (tgl)
  {
    /* OpenGL don't like odd textures... */
    if (tgl->width & 0x01 || tgl->height & 0x01)
    {
      fprintf (stderr, READPNG_EEPRFX"texture have odd dimentions (%dx%d)\n",
               tgl->width, tgl->height);
    }
    else
    {
      /* generate texture */
      glGenTextures (1, &texid);
      glBindTexture (GL_TEXTURE_2D, texid);
    
      /* setup some parameters for texture filters and mipmapping */
      glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
      glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      
      gluBuild2DMipmaps (GL_TEXTURE_2D, tgl->internalFormat,
                         tgl->width, tgl->height,
                         tgl->format, GL_UNSIGNED_BYTE, tgl->texels);
    }
    
    /* OpenGL has its own copy of texture data; we can safely delete our own */
    readpng_tgl_free (tgl);
  }
  
  return texid;
}
