/*
 *  Jomar3D, a lobster-running-to-catch game.
 *  Copyright (C) 2006-2008 Jonathan Michalon <studios.chalmion@no-log.org>.
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <GL/glut.h>

#include "core.h"
#include "display.h"
#include "drawtext.h"
#include "interface.h"
#include "jomar3D.h"
#include "scores.h"
#include "utils.h"

/* Maximum number of character a name is allowed to be built of */
#define MAXCNAME 50


/* if not compiling with GNUC, elide __attribute__ */
#if !defined (__GNUC__) && !defined (__attribute__)
# define __attribute__(x) /* nothing */
#endif


/* this function always returns the same value */
static size_t scores_nCharInt (void) __attribute__ ((const));
static int    scores_setName  (unsigned int number,
                               const char* name);


static char   l_name[MAXCNAME];     /* name beeing built by the player */
static Player l_players[MAXNAMES];  /* definition of each player */
static int    l_timeneeded;         /* time needed to win */
static int    l_stepsneeded;        /* steps needed to win */

static int l_validate_next_redisplay = 0; /* whether we should validate the name during next redisplay */


/* Init the pointers of "players" */
void
scores_init (void)
{
  int i;
  for (i=0; i<MAXNAMES; i++)
    l_players[i].name = NULL;
}

/* give the name of the player going to save his scores */
char*
scores_getCurrentName (void)
{
  return l_name;
}

/* give the name of a player */
/** Gets the name of a player from its ID
 * @param number  player ID
 * @returns a pointer to the player's name, or NULL if @param number is not
 *          a valid player ID.
 */
char*
scores_getName (unsigned int number)
{
  if (number >= MAXNAMES)
    return NULL;
  
  return l_players[number].name;
}

/** Sets the name of a player from its ID
 * @param number  player ID
 * @param name    name to set to the specified player
 * @returns 0 on success, -1 otherwise.
 * 
 * @warning @p name MUST has at least MAXCNAME characters.
 */
static int
scores_setName (unsigned int number, const char* name)
{
  Player* player = scores_getPlayer (number);
  
  if (!player)
  {
    fprintf (stderr, "Invalid player ID %d\n", number);
    return -1;
  }
  
  free (player->name);
  
  if (name == NULL)
    player->name = NULL;
  else
  {
    char* tmpname;
    /* we're not using strlen(name) because it ends with the number of played
     * games and an EOL... and probably all the score file
    size_t tmplen = strlen (name);
     */
    size_t tmplen = MAXCNAME;
    
    #define ALLOCCHECK(v) \
      do { \
        if (!v) { \
          perror ("failed to allocate memory"); \
          free (tmpname); \
          return -1; \
        } \
      } while (0)
    
    /* temporary copy of the name */
    tmpname = malloc (tmplen + 1);
    ALLOCCHECK (tmpname);
    strncpy (tmpname, name, tmplen + 1);
    
    /* remove trailing spaces at end of player name */
    while (tmpname[--tmplen] == ' ');
    tmpname[tmplen + 1] = '\0';
    
    /* real player name */
    player->name = malloc (strlen (tmpname) + 1);
    ALLOCCHECK (player->name);
    
    /* then copy clean name into player's name */
    strcpy (player->name, tmpname);
    
    free (tmpname);
    
    #undef ALLOCCHECK
  }
  
  return 0;
}

/** Gets all info about a player
 * @param number  player ID
 * @returns a pointer to the desired structure
 */
/* give the scores of a whole player */
Player*
scores_getPlayer (unsigned int number)
{
  if (number >= MAXNAMES)
    return NULL;
  
  return &l_players[number];
}

/* load all scores from file */
int
scores_load (void)
{
  FILE *file;
  int ngamers, i, j;
  char *tmpget;
  int ncharmaxint = scores_nCharInt (); /* number of chars needed for the greater int */
  
  /* Free all pointers to names of teh players */
  for (i=0; i<MAXNAMES; i++)
    l_players[i].name = NULL;
  
  tmpget = malloc (max ((ncharmaxint + 1) * NSTORE + 2,
                        MAXCNAME + ncharmaxint + 2));
  if (!tmpget)
  {
    perror ("Failed to allocate memory");
    return 0;
  }
  
  if (!(file = fopen (SCOREFILE, "r")))
  {
    perror ("fopen for read");
    fprintf (stderr, "Cannot open %s... no load of the scores!\n", SCOREFILE);
    free (tmpget);
    return 0;
  }

  ngamers = 0;
  for (i = 0;; i++)
  {
    /* get the next record */
    fgets (&tmpget[0],
           max ((ncharmaxint + 1) * NSTORE + 2, MAXCNAME + ncharmaxint + 2),
           file);
    if (i > 3 * MAXNAMES || feof (file))
    {
      fclose (file);
      free (tmpget);
      return ngamers;
    }
    /* store the name and the number of games done */
    if ((float) (i - 1) / 3. == (float) ((i - 1) / 3))
    {
      unsigned int id = ((i - 1) / 3);
      
      ngamers++;
      
      scores_setName (id, &tmpget[0]);
      l_players[id].games = atoi (&tmpget[MAXCNAME]);
    }
    /* store the times needed*/
    if ((float) (i - 2) / 3. == (float) ((i - 2) / 3))
    {
      for (j = 0; j < NSTORE; j++)
      {
        l_players[(i - 2) / 3].times[j] = atoi (&tmpget[j * (ncharmaxint + 1)]);
      }
    }
    /* store the steps needed */
    if ((float) (i - 3) / 3. == (float) ((i - 3) / 3) && i != 0)
    {
      for (j = 0; j < NSTORE; j++)
      {
        l_players[(i - 3) / 3].steps[j] = atoi (&tmpget[j * (ncharmaxint + 1)]);
      }
    }
  }
  
  fclose (file);
  free (tmpget);
  return ngamers;
}

/* write the scores into the file */

static void
scores_write (void)
{
  FILE *file;
  int i, j;
  int ncharmaxint = scores_nCharInt (); /* number of chars needed for the greater int */

  if (!(file = fopen (SCOREFILE, "w")))
  {
    perror ("fopen for write");
    fprintf (stderr, "Cannot open %s... no save of the scores!\n", SCOREFILE);
    return;
  }
  fprintf (file,
           "##### This is the storage file of the scores of jomar3D. #####\n");
  for (i = 0; i < MAXNAMES; i++)
  {
    if (l_players[i].name == NULL)
      break;
    fprintf (file, "%-*s %d\n", MAXCNAME, l_players[i].name, l_players[i].games);
    for (j = 0; j < NSTORE; j++)
      fprintf (file, "%*i ", ncharmaxint, l_players[i].times[j]);
    fprintf (file, "\n");
    for (j = 0; j < NSTORE; j++)
      fprintf (file, "%*i ", ncharmaxint, l_players[i].steps[j]);
    fprintf (file, "\n");
  }
  fclose (file);
}

/* should be registered as glut special key press callback after scores validating
 * change to main menu if it had not be done exactly at the same moment */
static void
scores_onSKeyDown (int c, int x, int y)
{
  (void) c, (void) x, (void) y;
  if (core_checkState (JOMAR3D_SCORE))
    interface_changeMenu(MP);
}

/* should be registered as glut normal key press callback after scores validating
 * change to main menu if enter or quit if escape */
static void
scores_onKeyDown (unsigned char c, int x, int y)
{
  (void) x, (void) y;
  if (c == 27) /* escape */
    exit (EXIT_SUCCESS);
  if (core_checkState (JOMAR3D_SCORE) && c == 13 /* enter */)
    interface_changeMenu(MP);
}

/* should be registered as glut timer callback after scores validating
 * change to main menu if we are in the scores interface
 * otherwise it had already be done by the player and we mustn't */
static void
scores_leave (int dummy)
{
  (void) dummy;
  if (core_checkState (JOMAR3D_SCORE))
    interface_changeMenu(MP);
}

/* validate the current entered name as given if it is the moment
 * or return 0 */
int
scores_validateName (void)
{
  int i, j, k, l;
  int m;
  int place = 0;

  if (!l_validate_next_redisplay)
    return 0;
  
  debugmsg (("Values to compare: %i ticks %i steps.\n", l_timeneeded, l_stepsneeded));

  for (i = 0; i < MAXNAMES; i++)
  {
    if (l_players[i].name != NULL)
    {
      if (strncmp (l_name, l_players[i].name, MAXCNAME) == 0)
      {
        debugmsg (("Known name.\n"));
        
        drawtext_drawString (750., 1400., 0.3, 0.3, TEXT_CENTER, 1400., FALSE, TRUE,
                             core_getTextPointer(34));
        drawtext_drawString (750., 1000., 0.3, 0.3, TEXT_CENTER, 1400., FALSE, TRUE,
                             core_getTextPointer(37));
        l_players[i].games += 1;
        for (j = 0; j < NSTORE; j++)
        {
          if (l_players[i].times[j] == 0)
          {
            debugmsg (("Not enough stat in times.\n"));
            
            drawtext_drawString (750., 1300., 1.0, 0.5, TEXT_CENTER, 1400., FALSE, TRUE,
                                 core_getTextPointer(33));
            l_players[i].times[j] = l_timeneeded;
            goto next;
          }
        }
        for (j = 0; j < NSTORE; j++)
        {
          if (l_players[i].times[j] > l_timeneeded)
          {
            debugmsg (("Better value in times.\n"));
            
            drawtext_drawString (750., 1300., 1.0, 0.5, TEXT_CENTER, 1400., FALSE, TRUE,
                                 core_getTextPointer(35));
            for (k = 0, l = 0, m = 0; k < NSTORE; k++)
            {
              if (m < l_players[i].times[k])
              {
                m = l_players[i].times[k];
                l = k;
              }
            }
            l_players[i].times[l] = l_timeneeded;
            goto next;
          }
        }
        
        debugmsg (("Nothing to say for time.\n"));
        
        drawtext_drawString (750., 1300., 1.0, 0.5, TEXT_CENTER, 1400., FALSE, TRUE,
                             core_getTextPointer(36));
      next:;
        for (j = 0; j < NSTORE; j++)
        {
          if (l_players[i].steps[j] == 0)
          {
            debugmsg (("Not enough stat in steps.\n"));
            
            drawtext_drawString (750., 900., 1.0, 0.5, TEXT_CENTER, 1400., FALSE, TRUE,
                                 core_getTextPointer(33));
            l_players[i].steps[j] = l_stepsneeded;
            goto write;
          }
        }
        for (j = 0; j < NSTORE; j++)
        {
          if (l_players[i].steps[j] > l_stepsneeded)
          {
            debugmsg (("Better value in steps.\n"));
            
            drawtext_drawString (750., 900., 1.0, 0.5, TEXT_CENTER, 1400., FALSE, TRUE,
                                 core_getTextPointer(38));
            for (k = 0, l = 0, m = 0; k < NSTORE; k++)
              if (m < l_players[i].steps[k])
              {
                m = l_players[i].steps[k];
                l = k;
              }
            l_players[i].steps[l] = l_stepsneeded;
            goto write;
          }
        }
        
        debugmsg (("Nothing to say for step.\n"));
        
        drawtext_drawString (750., 900., 1.0, 0.5, TEXT_CENTER, 1400., FALSE, TRUE,
                             core_getTextPointer(39));
        goto write;
      }
      else
        continue;
    }
    debugmsg (("New name.\n"));
    
    drawtext_drawString (750., 700., 1.0, 1.0, TEXT_CENTER, 1400., FALSE, TRUE,
                         core_getTextPointer(31));
    
    scores_setName (i, l_name);
    l_players[i].games = 1;
    l_players[i].times[0] = l_timeneeded;
    for (j = 1; j < NSTORE; j++)
      l_players[i].times[j] = 0;
    l_players[i].steps[0] = l_stepsneeded;
    for (j = 1; j < NSTORE; j++)
      l_players[i].steps[j] = 0;
    place = 1;
    break;
  }
  if (!place)
    printf ("No place more for a new name in the tabular - grow MAXNAMES.\n");

write:;
  glutKeyboardFunc (NULL);
  /*glutSpecialFunc (NULL); easier when not taking care of urging players */
  glutSpecialFunc (scores_onSKeyDown);
  glutKeyboardFunc (scores_onKeyDown);
  glutSpecialUpFunc (NULL);
  glutMouseFunc (NULL);
  glutMotionFunc (NULL);
  glutPassiveMotionFunc (NULL);
#if (MYFREEGLUT)
  glutMouseWheelFunc (NULL);
#endif
  /* glutTimerFunc (8000, interface_changeMenu, MP); without key press support*/
  glutTimerFunc (8000, scores_leave, 0);
  l_validate_next_redisplay = 0;
  scores_write ();
  return 1;
}

/* keyboard function for the score interface */

static void
scores_keyboardHandler (unsigned char c, int x, int y)
{
  size_t namelen = strlen (l_name);
  
  (void) x, (void) y;
  switch (c)
  {
    case 27:                   /* Escape */
      exit (EXIT_SUCCESS);
      break;
    case 13:                   /* Validate */
      if (namelen != 0)
        l_validate_next_redisplay = 1;
      break;
  }
  
  if (c == '\b')
  {
    l_name[namelen - 1] = '\0';
    l_name[namelen] = 0;
  }
  else
  {
    if (c > 31 && c < 128)   /* select the stored caracters */
      if (namelen < MAXCNAME)
        strncat (l_name, (char *) &c, 1);
  }
  glutPostRedisplay ();
  
  debugmsg (("%s | %c %d\n", l_name, c, c));
}

/* mouse interface for the scores (select name from list) */

static void
scores_mouseHandler (int button, int state, int x, int y)
{
  float xd, yd;
  int playernb = -1;

  xd = display_fitLaps (x, 1500, 1);
  yd = 1500. - display_fitLaps (y, 1500, 2);

  (void) button;
  
  debugmsg (("xd, yd %f %f\n", xd, yd));
  
  if (state != GLUT_DOWN)
    return;
  if (xd > 1400. || xd < 100.)
    return;
  if (yd > 1050. || yd < 300.)
    return;
  playernb = (int) ((1050. - yd) / (700. / MAXNAMES));
  
  debugmsg (("selected player number %d\n", playernb));
  
  if (l_players[playernb].name == NULL)
    return;
  strncpy (l_name, l_players[(int) ((1050. - yd) / (700. / MAXNAMES))].name,
           MAXCNAME);
  scores_keyboardHandler (13, 0, 0);
}

/* get the number of characters needed by the greatest integer number */
static size_t
scores_nCharInt (void)
{
  /* defines are for optimisation (values are from GLibC's limits.h) */
  
#if ULONG_MAX == 18446744073709551615ul /* WORDSIZE = 64 */
  return 20u;
#elif ULONG_MAX == 4294967295ul /* WORDSIZE = 32 */
  return 10u;
#else /* other strange arch */
  static size_t len = 0;
  
  /* don't recompute length if already done */
  if (len != 0)
    return len;
  
  len = uintlen (ULONG_MAX, 10);
  
  debugmsg (("Length in characters of ULONG_MAX (%lu) is %d\n", ULONG_MAX, len));
  
  return len;
#endif
}

/* begin the score interface */

void
scores_begin (int time)
{
  int i;

  core_addState (JOMAR3D_SCORE);
  l_timeneeded = time;
  l_stepsneeded = core_getMartianSteps ();
  /* ensure the lobster stops otherwise it will continue to run
   * after the score interface */
  interface_onKeyUp (GLUT_KEY_UP, 0, 0);
  glutKeyboardFunc (scores_keyboardHandler);
  glutKeyboardUpFunc (NULL);
  glutSpecialFunc (NULL);
  glutSpecialUpFunc (NULL);
  glutMouseFunc (scores_mouseHandler);
  scores_load ();
  for (i = 0; i < MAXCNAME; i++)
    l_name[i] = 0;
  l_validate_next_redisplay = 0;
  glutPostRedisplay ();
}
